def get_first_shared_inheritance(classes: list):
    """
    returns the type all classes inherit from.
    supports multiple inheritance.
    """

    if len(classes) == 0:
        return None

    if len(classes) == 1:
        return classes[0]

    first_shared_inheritance = classes[0]

    for cls in classes[1:]:
        first_shared_inheritance = _get_first_shared_inheritance_recursive(
            cls, first_shared_inheritance
        )

    return first_shared_inheritance


def _get_first_shared_inheritance_recursive(class_a, class_b):
    """
    returns the type both classes inherit from.
    """

    if class_a == class_b:
        return class_a

    if class_a in class_b.__bases__:
        return class_a

    if class_b in class_a.__bases__:
        return class_b

    for base_type in class_a.__bases__:
        result = _get_first_shared_inheritance_recursive(base_type, class_b)

        if result is not None:
            return result

    for base_type in class_b.__bases__:
        result = _get_first_shared_inheritance_recursive(class_a, base_type)

        if result is not None:
            return result

    return None


import inspect
from typing import Any, Tuple, Type, TypeVar
from ..mutable._mutable import Mutable
from ._operators import *
from inspect import signature
from ..assertion import assert_type, assert_true
from ..variables import get_names


T = TypeVar("T")


class Link(Mutable[T]):
    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    def __add__(self, other):
        return Add(self, other)

    def __radd__(self, other):
        return Add(other, self)

    def __sub__(self, other):
        return Sub(self, other)

    def __rsub__(self, other):
        return RSub(self, other)

    def __mul__(self, other):
        return Mul(self, other)

    def __rmul__(self, other):
        return Mul(other, self)

    def __truediv__(self, other):
        return Div(self, other)

    def __rtruediv__(self, other):
        return RDiv(self, other)

    def __pow__(self, other):
        return Pow(self, other)

    def __rpow__(self, other):
        return RPow(self, other)

    def __mod__(self, other):
        return Mod(self, other)

    def __rmod__(self, other):
        return Mod(other, self)

    def __abs__(self):
        return Abs(self)

    def __neg__(self):
        return Mul(self, -1)

    def __invert__(self):
        return Not(self)

    def __eq__(self, other):
        return Eq(self, other)

    def __ne__(self, other):
        return Ne(self, other)

    def __gt__(self, other):
        return Gt(self, other)

    def __ge__(self, other):
        return Ge(self, other)

    def __lt__(self, other):
        return Lt(self, other)

    def __le__(self, other):
        return Le(self, other)

    def __and__(self, other):
        return And(self, other)

    def __rand__(self, other):
        return And(other, self)

    def __or__(self, other):
        return Or(self, other)

    def __ror__(self, other):
        return Or(other, self)

    def __xor__(self, other):
        return Xor(self, other)

    def __rxor__(self, other):
        return Xor(other, self)

    def __iadd__(self, other):
        other_value = other.value if isinstance(other, OperatorLink) else other
        self.value += other_value
        return self


class Attr(Link[T]):
    def __init__(self, obj: Any, attr: str, type_hint: Type[T] = Any, immutable=False):
        self.obj = obj
        self.attr_list = attr.split(".")
        self.immutable = immutable
        obj_names = get_names(obj, inspect.currentframe().f_back)
        obj_repr = repr(obj)
        has_no_repr = obj_repr.startswith("<") and obj_repr.endswith(">")
        # try to avoid using builtin <Class object at 0x2974c29d8f> repr by getting names in frame back
        self.obj_name = obj_names[0] if len(obj_names) > 0 and has_no_repr else obj_repr

    @property
    def attr_holder(self):
        attr_holder = self.obj.value if isinstance(self.obj, Mutable) else self.obj

        for attr in self.attr_list[:-1]:
            attr_holder = getattr(attr_holder, attr)
        
        return attr_holder
    
    @property
    def final_attr(self):
        return self.attr_list[-1]

    @property
    def value(self) -> T:
        return getattr(self.attr_holder, self.final_attr)

    @value.setter
    def value(self, value: T):
        if self.immutable:
            raise AttributeError(
                "attribute is immutable. modify it using set_value(value)."
            )
        setattr(self.attr_holder, self.final_attr, value)

    def set_value(self, value: T):
        setattr(self.attr_holder, self.final_attr, value)

    def __repr__(self):
        return f"Attr({self.obj_name}, \"{'.'.join(self.attr_list)}\")"


class Var(Link[T]):
    def __init__(self, value: T, immutable=False):
        self._value = value
        self.immutable = immutable

    @property
    def value(self) -> T:
        return self._value

    @value.setter
    def value(self, value: T):
        if self.immutable:
            raise AttributeError(
                "variable is immutable. modify it using set_value(value)."
            )
        self._value = value

    def set_value(self, value: T):
        self._value = value


class OperatorLink(Link[T]):
    def __init__(self, operator: Type[Operator], *values: "Tuple[T, ...]"):
        assert_true(issubclass(operator, Operator))
        assert_type(
            values[0],
            Mutable,
            f"the first input to an OperatorLink like {self.__class__.__name__} must be a Mutable. use Var(value) to create one.",
        )

        self.operator: Type[Operator] = operator
        self.inputs = list(values)

        values_given = len(values)
        values_expected = signature(operator._eval).parameters.__len__()
        if values_given != values_expected:
            for attr in signature(operator._eval).parameters.values():
                # If an unpack operation is used, the given arguments will not exceed the expected arguments.
                if str(attr).startswith("*") and not str(attr).startswith("**"):
                    break
            else:
                raise ValueError(
                    f"wrong number of values for {operator.__name__} (expected {values_expected}, got {values_given})"
                )

        self.order = operator.order
        self.print_pattern = operator.print_pattern

    @property
    def value(self) -> T:
        inputs = (
            value.value if isinstance(value, Mutable) else value
            for value in self.inputs
        )

        return self.operator(*inputs)

    @value.setter
    def value(self, value):
        # Modify the first value using the reverse operation
        try:
            value1 = self.operator.reverse(value, *self.inputs[1:])
            if value1 is not None:
                if isinstance(self.inputs[0], Mutable):
                    self.inputs[0].__set__(self, value1)
                else:
                    self.inputs[0] = value1
            else:
                raise ValueError(
                    f"value cannot be set: {self.operator.__name__}({self.inputs[0]}, <Any>) != {value}"
                )
        except NotImplementedError as e:
            print(e.with_traceback())
            print("cannot set value for link without reverse operation")

    def __repr__(self):
        # return the link string using the print pattern
        return_value = self.print_pattern
        operator_inputs = signature(self.operator._eval).parameters

        for index, (value, name) in enumerate(zip(self.inputs, operator_inputs)):

            def is_paren_needed(value):
                if len(operator_inputs) <= 1:
                    return False
                if isinstance(value, OperatorLink):
                    if value.order < self.order:
                        return True
                    if (
                        index != 0
                        and value.order == self.order
                        and value.operator.__name__ != self.operator.__name__
                    ):
                        return True

            if is_paren_needed(value):
                value = f"({repr(value)})"

            else:
                value = repr(value)

            return_value = return_value.replace(f"${name}", value)

        return return_value

    def __getitem__(self, key):
        return self.inputs[key]

    def __setitem__(self, key, value):
        self.inputs[key] = value


class Eq(OperatorLink[bool]):
    def __init__(self, a: OperatorLink, b: OperatorLink):
        super().__init__(EqualOperator, a, b)


class Ne(OperatorLink[bool]):
    def __init__(self, a: OperatorLink, b: OperatorLink):
        super().__init__(NotEqualOperator, a, b)


class Gt(OperatorLink[bool]):
    def __init__(self, a: OperatorLink, b: OperatorLink):
        super().__init__(GreaterOperator, a, b)


class Ge(OperatorLink[bool]):
    def __init__(self, a: OperatorLink, b: OperatorLink):
        super().__init__(GreaterEqualOperator, a, b)


class Lt(OperatorLink[bool]):
    def __init__(self, a: OperatorLink, b: OperatorLink):
        super().__init__(LessOperator, a, b)


class Le(OperatorLink[bool]):
    def __init__(self, a: OperatorLink, b: OperatorLink):
        super().__init__(LessEqualOperator, a, b)


class NumberOperatorLink(OperatorLink[T]):
    pass


class Add(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(AdditionOperator, a, b)


class Sub(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(SubtractionOperator, a, b)


class RSub(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsSubtractionOperator, a, b)


class Mul(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(MultiplicationOperator, a, b)


class Div(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(DivisionOperator, a, b)


class RDiv(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsDivisionOperator, a, b)


class Pow(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(PowerOperator, a, b)


class RPow(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsPowerOperator, a, b)


class Root(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(RootOperator, a, b)


class RRoot(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsRootOperator, a, b)


class Mod(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(ModuloOperator, a, b)


class Abs(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T"):
        super().__init__(AbsoluteOperator, a)


class Ln(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T"):
        super().__init__(NaturalLogarithmOperator, a)


class LogB(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BaseBLogarithmOperator, a, b)


class RLogB(NumberOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(BackwardsBaseBLogarithmOperator, a, b)


class BooleanOperatorLink(OperatorLink[T]):
    pass


class And(BooleanOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(AndOperator, a, b)


class Or(BooleanOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(OrOperator, a, b)


class Xor(BooleanOperatorLink[T]):
    def __init__(self, a: "Link[T] | T", b: "Link[T] | T"):
        super().__init__(XorOperator, a, b)


class Not(BooleanOperatorLink[T]):
    def __init__(self, a: "Link[T] | T"):
        super().__init__(NotOperator, a)



def list(list, final_sep="and", empty_str="nothing"):
    """
    convert a list to a string with a final separator
    """

    if len(list) == 0:
        return empty_str
        
    string = str(list[-1])

    if len(list) > 1:
        string = f"{str(list[-2])} {final_sep} {string}"
    
    for i in range(3, len(list) + 1):
        string = f"{str(list[-i])}, {string}"
    
    return string


import os
from typing import Any
from ..constants import UNDEFINED


def load_file(file_path: str, create_if_nonexistant=True, default: Any = UNDEFINED):
    if not isinstance(file_path, str):
        if default is UNDEFINED:
            raise TypeError("file_path must be a string")
        return default
    if create_if_nonexistant and not os.path.exists(file_path):
        return open(file_path, "x")
    else:
        try:
            return open(file_path, "w")
        except FileNotFoundError as e:
            if default is UNDEFINED:
                raise e
            return default


from typing import Any, Callable, Dict, Iterable, Tuple, Type

import inspect
from ..debug import Log
from .. import to_string, get_first_shared_inheritance, get_types


Log.load_file("testing.log")


PASSED = f"[style: bright, color: green]PASSED[style: reset_all]"
FAILED =   f"[style: bright, color: red]!!! FAILED[style: reset_all]"
SUM_PASSED = f"[style: bright, color: green]PASSED[style: reset_all]"
SUM_FAILED =   f"[style: bright, color: red]FAILED[style: reset_all]"


def _format_function(string):
    return f"[style: bright, color: magenta]{string}[style: reset_all]"

def _format_object(string):
    return f"[style: bright, color: cyan]{string}[style: reset_all]"


class TestSummary:
    def __init__(self, object_type: type, functions: Iterable[Callable], total_fails: int, total_tests: int):
        self.object_type = object_type
        self.functions = functions
        self.total_fails = total_fails
        self.total_tests = total_tests
    
    @property
    def function_name_string(self) -> Iterable[str]:
        return to_string.list(
            [_format_function(function.__name__.replace("_", " ")) for function in self.functions]
        )
    
    @property
    def object_name_string(self) -> str:
        return _format_object(self.object_type.__name__)
    
    @property
    def raw_function_name_string(self) -> Iterable[str]:
        return to_string.list(
            [function.__name__.replace("_", " ") for function in self.functions]
        )
    
    @property
    def raw_object_name_string(self) -> str:
        return self.object_type.__name__



def test(objects, *assert_functions, iterations=1):
    """
    test a list of objects with one or more given assert functions
    """

    if isinstance(objects, Callable):
        assert_functions = (*assert_functions, 
        objects)

    for func in assert_functions:
        func.takes_index = func.__code__.co_argcount > 1
        func.takes_object = func.__code__.co_argcount > 0


    assert_function_strings = []

    max_raw_function_string_length = 0

    for func in assert_functions:
        assert_function_strings.append(
            _format_function(func.__name__.replace('_', ' '))
        )
        max_raw_function_string_length = max(max_raw_function_string_length, len(func.__name__))

    object_strings = []

    max_raw_object_string_length = 0

    for object_ in objects:
        if hasattr(object_, "name"):
            object_string = object_.name
        elif hasattr(object_, "__name__"):
            object_string = object_.__name__
        else:
            object_string = repr(object_)
        
        max_raw_object_string_length = max(len(object_string), max_raw_object_string_length)
        
        object_string = _format_object(object_string)
        
        object_strings.append(object_string)

    max_object_string_length = max(len(object_string) for object_string in object_strings)

    max_function_string_length = max(len(string) for string in assert_function_strings)
    
    summary_string = f"\n\n\n{' ' * int((max_raw_object_string_length + max_raw_function_string_length + 15) / 2)}--- summary ---\n\n"

    total_fails = 0

    for object_, object_string in zip(objects, object_strings):

        print("")
        
        for assert_function, assert_function_string in zip(assert_functions, assert_function_strings):
            fails = 0
            for i in range(iterations):
                return_string = ""
                try:
                    if assert_function.takes_index:
                        return_string = assert_function(object_, i)
                    elif assert_function.takes_object:
                        return_string = assert_function(object_)
                    else:
                        return_string = assert_function()
                except AssertionError as e:
                    Log(f"{FAILED}: {object_string} - {assert_function_string} - assertion failed: {e}")
                    fails += 1
                    total_fails += 1
                    continue
                except Exception as e:
                    Log(f"{FAILED}: {object_string} - {assert_function_string} - exception thrown: {e}")
                    fails += 1
                    total_fails += 1
                    continue
                if not isinstance(return_string, str):
                    return_string = ""
                else:
                    return_string = f" with message: {return_string}"
                Log(f"{PASSED}: {object_string} - {assert_function_string} - passed{return_string}")
            if fails == 0:
                summary_string += f"{SUM_PASSED}: {object_string: <{max_object_string_length}} - {assert_function_string: <{max_function_string_length}} - passed all tests\n"
            else:
                summary_string += f"{SUM_FAILED}: {object_string: <{max_object_string_length}} - {assert_function_string: <{max_function_string_length}} - failed {fails}/{iterations} tests\n"
    
    Log(summary_string)

    return TestSummary(get_first_shared_inheritance(get_types(*objects)), assert_functions, total_fails, iterations * len(objects) * len(assert_functions))


def multi_test(*arguments: Tuple[Tuple[Iterable[Any], Callable, Callable, Dict[str, Any]]]):
    """
    run multiple tests and recieve a summary of the results.
    syntax: 
    ```python
    multi_test(
        (objects_1, function_1, function_2, {'iterations': iterations}),
        (objects_2, function_1, function_2, {'iterations': iterations}),
    )
    ```
    """

    summaries: list[TestSummary] = []

    for arguments_ in arguments:
        args = arguments_ if not isinstance(arguments_[-1], dict) else arguments_[:-1]
        kwargs = arguments_[-1] if isinstance(arguments_[-1], dict) else {}
        summaries.append(test(*args, **kwargs))

    longest_object_string = max(len(summary.object_name_string) for summary in summaries)

    longest_raw_object_string = max(len(summary.raw_object_name_string) for summary in summaries)
    longest_raw_function_string = max(len(summary.raw_function_name_string) for summary in summaries)

    summary_strings = []

    for summary in summaries:
        if summary.total_fails == 0:
            summary_strings.append(
                f"{SUM_PASSED}: {summary.object_name_string: <{longest_object_string}}"
                f" - {summary.function_name_string}"
                f"{' ' * (longest_raw_function_string - len(summary.raw_function_name_string))}"
                f" - passed all tests")
        else:
            summary_strings.append(
                f"{SUM_FAILED}: {summary.object_name_string: <{longest_object_string}}"
                f" - {summary.function_name_string}"
                f"{' ' * (longest_raw_function_string - len(summary.raw_function_name_string))}"
                f" - failed {summary.total_fails}/{summary.total_tests} tests")

    summary = f"\n\n\n{' ' * int((longest_raw_object_string + longest_raw_function_string + 15) / 2)}--- summary ---\n\n"
    summary += "\n".join(summary_strings)

    Log(summary)

    return summaries


def test_from_class(cls: Type[Any], iterations=1) -> TestSummary:
    """
    run a test with all functions in a class that start with `test_`
    """

    return test([None], *(func for name, func in inspect.getmembers(cls, predicate=inspect.isfunction) if name.startswith("test_")))


from typing import Iterable, Iterator, SupportsIndex, Tuple, TypeVar


T = TypeVar("T")


class EList(list, Iterable[T]):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and args[0] is None:
            super().__init__()
        else:
            super().__init__(*args, **kwargs)
        
        self._iterators: list[EListIterator] = []
        
    def insert(self, __index: SupportsIndex, __object: T) -> None:
        for iterator in self._iterators:
            if iterator.index >= __index:
                iterator.index += 1

        super().insert(__index, __object)
        return self
    
    def remove(self, __value: T) -> None:
        index = self.index(__value)
        self.pop(index)
        return self
    
    def pop(self, __index: SupportsIndex = ...) -> T:
        for iterator in self._iterators:
            if iterator.index >= __index:
                iterator.index -= 1

        super().pop(__index)
        return self

    def merge(self, other):
        for item in other:
            if item not in self:
                self.append(item)
        return self
    
    def merged(self, other):
        new_list = self.copy()
        new_list.merge(other)
        return new_list

    def clear_duplicates(self):
        original_list = self.copy()
        self.clear()
        for item in original_list:
            if item not in self:
                self.append(item)
        return self

    def without_duplicates(self):
        new_list = EList()
        for item in self:
            if item not in new_list:
                new_list.append(item)
        return new_list
    
    def _clear_from_single_obj(self, object):
        '''
        clear the list from the given object.
        '''
        while True:
            try:
                self.remove(object)
            except ValueError:
                return
    
    def clear_from(self, *objects):
        '''
        remove all given objects from the list.
        '''
        for object in objects:
            self._clear_from_single_obj(object)
        return self
        
    def without(self, *objects):
        '''
        return a new list without the given objects.
        '''
        new_list = self.copy()
        new_list.clear_from(*objects)
        return new_list
    
    def extend(self, __iterable: Iterable[T]) -> None:
        super().extend(__iterable)
        return self

    def extended(self, iterable):
        new_list = self.copy()
        new_list.extend(iterable)
        return new_list
        
    def __add__(self, other):
        new_list = self.copy()
        if isinstance(other, Iterable):
            new_list.extend(other)
        else:
            new_list.append(other)
        return new_list
    
    def __radd__(self, other):
        return self + other
    
    def __iadd__(self, other):
        if isinstance(other, Iterable):
            self.extend(other)
        else:
            self.append(other)
        return self
    
    def __sub__(self, other):
        if isinstance(other, Iterable):
            return self.without(*other)
        else:
            return self.without(other)

    def __rsub__(self, other):
        return EList(other) - self

    def __isub__(self, other):
        if isinstance(other, Iterable):
            self.clear_from(*other)
        else:
            self.clear_from(other)
        return self
    
    def __getitem__(self, index) -> T:
        return super().__getitem__(index)
    
    def __iter__(self) -> "EListIterator[T]":
        return EListIterator(self)
    
    def __and__(self, other):
        new_list = EList()
        for object in self:
            if object in other and object not in new_list:
                new_list.append(object)
        
        return new_list
    
    def __rand__(self, other):
        return self & other
    
    def __iand__(self, other):
        for object in self:
            if object not in other:
                self.remove(object)
        
        return self
    
    def __or__(self, other):
        return self.merged(other)
    
    def __ror__(self, other):
        return self | other
    
    def __ior__(self, other):
        self.merge(other)
        return self

    def __xor__(self, other):
        new_list = EList()
        for object in self:
            if object not in other and object not in new_list:
                new_list.append(object)
        for object in other:
            if object not in self and object not in new_list:
                new_list.append(object)
        
        return new_list
    
    def __rxor__(self, other):
        return self ^ other
    
    def __ixor__(self, other):
        self[:] = (self ^ other)[:]
    
    def enumerate(self):
        return EnumeratedEListIterator(self)

    def copy(self) -> "EList[T]":
        return EList(self)

    def join(self, sep: str = " ") -> str:
        '''
        join elements of the list together. all elements must be of type str.
        '''
        return sep.join([str(item) for item in self])


class EListIterator(Iterator[T]):
    def __init__(self, e_list: EList[T]):
        self.e_list = e_list
        self.e_list._iterators.append(self)
        self.index = 0
    
    def __iter__(self):
        return self
    
    def __next__(self) -> T:
        if self.index >= len(self.e_list):
            self.e_list._iterators.remove(self)
            raise StopIteration
        else:
            self.index += 1
            return self.e_list[self.index - 1]
    

class EnumeratedEListIterator(EListIterator, Iterator[Tuple[int, T]]):
    def __next__(self) -> T:
        if self.index >= len(self.e_list):
            self.e_list._iterators.remove(self)
            raise StopIteration
        else:
            self.index += 1
            return (self.index - 1, self.e_list[self.index - 1])
    